﻿# -*- coding: UTF-8 -*-
"""
   Copyright 2016 Esri
   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at
       http://www.apache.org/licenses/LICENSE-2.0
   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.​
    This sample adds and updates GNSS metadata fields in a feature class.
"""
import arcpy
import argparse
import os

def get_geodatabase_path(input_layer):
    """
	Gets the parent geodatabase of the layer
	:param input_layer: (string) The feature layer to get the parent database of
    :return: (string) The path to the geodatabase
    """
    workspace = os.path.dirname(arcpy.Describe(input_layer).catalogPath)
    if [any(ext) for ext in ('.gdb', '.mdb', '.sde') if ext in os.path.splitext(workspace)]:
        return workspace
    else:
        return os.path.dirname(workspace)

def check_and_create_domains(geodatabase):
    """
    Checks if the domains already exist, if they do
    then it checks the values and ranges

    If the domains do not exist, they are created

    :param geodatabase: (string) the path to the geodatabase to check
    :return:
    """
    domains = arcpy.da.ListDomains(geodatabase)
    domain_names = [domain.name for domain in domains]
    if 'ESRI_FIX_TYPE_DOMAIN' in domain_names:
        for domain in domains:
            if domain.name == 'ESRI_FIX_TYPE_DOMAIN':
                # check if cvs 0,1,2,4,5 are in the codedValues
                values = [cv for cv in domain.codedValues]
                if not set(set([0, 1, 2, 4, 5])).issubset(values):
                    arcpy.AddError("ESRI_FIX_TYPE_DOMAIN is missing a coded value pair.")
                    return
    else:
        # Add the domain and values
        arcpy.AddMessage('Adding ESRI_FIX_TYPE_DOMAIN domain to parent geodatabase...')
        
        arcpy.CreateDomain_management(in_workspace=geodatabase,
                                      domain_name="ESRI_FIX_TYPE_DOMAIN",
                                      domain_description="Fix Type",
                                      field_type="SHORT",
                                      domain_type="CODED",
                                      split_policy="DEFAULT",
                                      merge_policy="DEFAULT")

        arcpy.AddCodedValueToDomain_management(in_workspace=geodatabase,
                                               domain_name="ESRI_FIX_TYPE_DOMAIN",
                                               code="0",
                                               code_description="Fix not valid")
        arcpy.AddCodedValueToDomain_management(in_workspace=geodatabase,
                                               domain_name="ESRI_FIX_TYPE_DOMAIN",
                                               code="1",
                                               code_description="GPS")
        arcpy.AddCodedValueToDomain_management(in_workspace=geodatabase,
                                               domain_name="ESRI_FIX_TYPE_DOMAIN",
                                               code="2",
                                               code_description="Differential GPS")
        arcpy.AddCodedValueToDomain_management(in_workspace=geodatabase,
                                               domain_name="ESRI_FIX_TYPE_DOMAIN",
                                               code="4",
                                               code_description="RTK Fixed")
        arcpy.AddCodedValueToDomain_management(in_workspace=geodatabase,
                                               domain_name="ESRI_FIX_TYPE_DOMAIN",
                                               code="5",
                                               code_description="RTK Float")
    # Check if 'NumSats" is a domain, if so check the range
    if 'ESRI_NUM_SATS_DOMAIN' in domain_names:
        if domain.name == "ESRI_NUM_SATS_DOMAIN":
            if domain.range[0] != 0 or domain.range[1] != 99:
                arcpy.AddError("ESRI_NUM_SATS_DOMAIN domain has invalid range")
                return
    else:
        # Add the domain and set the range
        arcpy.AddMessage("Adding ESRI_NUM_SATS_DOMAIN to parent database...")
        arcpy.CreateDomain_management(in_workspace=geodatabase,
                                      domain_name="ESRI_NUM_SATS_DOMAIN",
                                      domain_description="Number of Satellites",
                                      field_type="SHORT",
                                      domain_type="RANGE",
                                      split_policy="DEFAULT",
                                      merge_policy="DEFAULT")
        arcpy.SetValueForRangeDomain_management(geodatabase, "ESRI_NUM_SATS_DOMAIN", 0, 99)
    # Check if 'StationID" is a domain, if so check the range
    if 'ESRI_STATION_ID_DOMAIN' in domain_names:
        if domain.name == "ESRI_STATION_ID_DOMAIN":
            if domain.range[0] != 0 or domain.range[1] != 1023:
                arcpy.AddError("ESRI_STATION_ID_DOMAIN domain has invalid range")
                return
    else:
        # Add the domain and set the range
        arcpy.AddMessage("Adding ESRI_STATION_ID_DOMAIN to parent database...")
        arcpy.CreateDomain_management(in_workspace=geodatabase,
                                      domain_name="ESRI_STATION_ID_DOMAIN",
                                      domain_description="Station ID",
                                      field_type="SHORT",
                                      domain_type="RANGE",
                                      split_policy="DEFAULT",
                                      merge_policy="DEFAULT")
        arcpy.SetValueForRangeDomain_management(geodatabase, "ESRI_STATION_ID_DOMAIN", 0, 1023)


def add_gnss_fields(feature_layer):
    """
    This adds specific fields required for Eos Tools Pro
        Laser offset to auto-populate collector application

        This will report errors if:
            1) Any of the fields already exist
            2) The input layer is not a point layer
            3) The layer is not found
            4) The layer is a shapefile

    Example: add_gps_fields(r"C:/temp/test.shp")

    :param feature_layer: (string) The feature layer (shapefile, feature class, etc) to add the fields to
    :return:
    """

    try:
        if not arcpy.Exists(feature_layer):
            arcpy.AddError("Feature layer: {} not found!".format(feature_layer))
            return
        if arcpy.Describe(feature_layer).shapeType != "Point":
            arcpy.AddError("Feature layer: {} is not a point layer".format(feature_layer))
            return
        if arcpy.Describe(feature_layer).dataType == "Feature Layer":
            if arcpy.Describe(feature_layer).dataElement.dataType == "ShapeFile":
                arcpy.AddError("ShapeFiles are not supported.")
                return
        elif arcpy.Describe(feature_layer).dataType == "ShapeFile":
            arcpy.AddError("ShapeFiles are not supported.")
            return

        # check if it's a service or feature class in db
        # Check the domains to see if they exist and are valid
        # will update if necessary
        
        if r'/rest/services' not in arcpy.Describe(feature_layer).catalogPath:
            arcpy.AddMessage('Checking domains...')
            geodatabase = get_geodatabase_path(feature_layer)
            check_and_create_domains(geodatabase)

        # Add the fields
        arcpy.AddMessage('Adding Required Fields...')
        
        # Add GNSS metadata fields
        existingFields = [field.name for field in arcpy.ListFields(feature_layer)]
        
        if 'ESRIGNSS_RECEIVER' not in existingFields:
            arcpy.AddField_management(feature_layer,
                                      'ESRIGNSS_RECEIVER',
                                      field_type="STRING",
                                      field_length=50,
                                      field_alias='Receiver Name',
                                      field_is_nullable="NULLABLE"
                                      )

        if 'EOSLASER_METHOD' not in existingFields:
            arcpy.AddField_management(feature_layer,
                                      'EOSLASER_METHOD',
                                      field_type="STRING",
                                      field_length=30,
                                      field_alias='Measurement Method',
                                      field_is_nullable="NULLABLE"
                                      )

        if 'ESRIGNSS_LATITUDE' not in existingFields:
            arcpy.AddField_management(feature_layer,
                                      'ESRIGNSS_LATITUDE',
                                      field_type="DOUBLE",
                                      field_alias='Offset Point Latitude',
                                      field_is_nullable="NULLABLE"
                                      )

        if 'ESRIGNSS_LONGITUDE' not in existingFields:
            arcpy.AddField_management(feature_layer,
                                      'ESRIGNSS_LONGITUDE',
                                      field_type="DOUBLE",
                                      field_alias='Offset Point Longitude',
                                      field_is_nullable="NULLABLE"
                                      )
                                      
        if 'ESRIGNSS_ALTITUDE' not in existingFields:
            arcpy.AddField_management(feature_layer,
                                      'ESRIGNSS_ALTITUDE',
                                      field_type="DOUBLE",
                                      field_alias='Offset Point HAE (ellipsoidal)',
                                      field_is_nullable="NULLABLE"
                                      )
                                                                         
        if 'EOS_ORTHO_HEIGHT' not in existingFields:
            arcpy.AddField_management(feature_layer,
                                      'EOS_ORTHO_HEIGHT',
                                      field_type="DOUBLE",
                                      field_alias='Orthometric Height',
                                      field_is_nullable="NULLABLE"
                                      )

        if 'EOS_UNDULATION' not in existingFields:
            arcpy.AddField_management(feature_layer,
                                      'EOS_UNDULATION',
                                      field_type="DOUBLE",
                                      field_alias='Altitude Undulation',
                                      field_is_nullable="NULLABLE"
                                      )

        if 'EOS_GEOID_MODEL' not in existingFields:
            arcpy.AddField_management(feature_layer,
                                      'EOS_GEOID_MODEL',
                                      field_type="STRING",
                                      field_length=50,
                                      field_alias='Geoid Model',
                                      field_is_nullable="NULLABLE"
                                      )
                                      
        if 'EOSLASER_DEVICE' not in existingFields:
            arcpy.AddField_management(feature_layer,
                                      'EOSLASER_DEVICE',
                                      field_type="STRING",
                                      field_length=50,
                                      field_alias='Bluetooth Laser Device',
                                      field_is_nullable="NULLABLE"
                                      )
                                      
        if 'EOSLASER_GNSSANTH' not in existingFields:
            arcpy.AddField_management(feature_layer,
                                      'EOSLASER_GNSSANTH',
                                      field_type="FLOAT",
                                      field_alias='GNSS Antenna Height',
                                      field_is_nullable="NULLABLE"
                                      )
                                      
        if 'EOSLASER_DEVICEH' not in existingFields:
            arcpy.AddField_management(feature_layer,
                                      'EOSLASER_DEVICEH',
                                      field_type="FLOAT",
                                      field_alias='Laser Device Height',
                                      field_is_nullable="NULLABLE"
                                      )
                                      
        if 'EOSLASER_MAGDEC' not in existingFields:
            arcpy.AddField_management(feature_layer,
                                      'EOSLASER_MAGDEC',
                                      field_type="FLOAT",
                                      field_alias='Magnetic Declination',
                                      field_is_nullable="NULLABLE"
                                      )
                                      
        if 'EOSLASER_CTL1_LAT' not in existingFields:
            arcpy.AddField_management(feature_layer,
                                      'EOSLASER_CTL1_LAT',
                                      field_type="FLOAT",
                                      field_alias='CP1 Latitude',
                                      field_is_nullable="NULLABLE"
                                      )
                                      
        if 'EOSLASER_CTL1_LON' not in existingFields:
            arcpy.AddField_management(feature_layer,
                                      'EOSLASER_CTL1_LON',
                                      field_type="FLOAT",
                                      field_alias='CP1 Longitude',
                                      field_is_nullable="NULLABLE"
                                      )
                                      
        if 'EOSLASER_CTL1_ALT' not in existingFields:
            arcpy.AddField_management(feature_layer,
                                      'EOSLASER_CTL1_ALT',
                                      field_type="FLOAT",
                                      field_alias='CP1 Altitude',
                                      field_is_nullable="NULLABLE"
                                      )
                                      
        if 'EOSLASER_CTL1_HRMS' not in existingFields:
            arcpy.AddField_management(feature_layer,
                                      'EOSLASER_CTL1_HRMS',
                                      field_type="FLOAT",
                                      field_alias='CP1 HRMS',
                                      field_is_nullable="NULLABLE"
                                      )
                                      
        if 'EOSLASER_CTL1_SATS' not in existingFields:
            arcpy.AddField_management(feature_layer,
                                      'EOSLASER_CTL1_SATS',
                                      field_type="SHORT",
                                      field_alias='CP1 Number of Sats',
                                      field_is_nullable="NULLABLE"
                                      )
                                      
        if 'EOSLASER_CTL1_FIXTYPE' not in existingFields:
            arcpy.AddField_management(feature_layer,
                                      'EOSLASER_CTL1_FIXTYPE',
                                      field_type="SHORT",
                                      field_alias='CP1 Fix Type',
                                      field_is_nullable="NULLABLE",
                                      field_domain = "ESRI_FIX_TYPE_DOMAIN"                                                                     )
                                      
        if 'EOSLASER_CTL1_AGE' not in existingFields:
            arcpy.AddField_management(feature_layer,
                                      'EOSLASER_CTL1_AGE',
                                      field_type="FLOAT",
                                      field_alias='CP1 Diff Age',
                                      field_is_nullable="NULLABLE"
                                      )
                                      
        if 'EOSLASER_CTL1_DIFFID' not in existingFields:
            arcpy.AddField_management(feature_layer,
                                      'EOSLASER_CTL1_DIFFID',
                                      field_type="STRING",
                                      field_length=4,
                                      field_alias='CP1 Station ID',
                                      field_is_nullable="NULLABLE"
                                      )
                                      
        if 'EOSLASER_CTL1_AVG' not in existingFields:
            arcpy.AddField_management(feature_layer,
                                      'EOSLASER_CTL1_AVG',
                                      field_type="SHORT",
                                      field_alias='CP1 Points Averaged',
                                      field_is_nullable="NULLABLE"
                                      )
                                      
        if 'EOSLASER_CTL1_SLDIST' not in existingFields:
            arcpy.AddField_management(feature_layer,
                                      'EOSLASER_CTL1_SLDIST',
                                      field_type="FLOAT",
                                      field_alias='CP1 Laser Slope Distance',
                                      field_is_nullable="NULLABLE"
                                      )
                                      
        if 'EOSLASER_CTL1_AZI' not in existingFields:
            arcpy.AddField_management(feature_layer,
                                      'EOSLASER_CTL1_AZI',
                                      field_type="FLOAT",
                                      field_alias='CP1 Laser Azimuth',
                                      field_is_nullable="NULLABLE"
                                      )
                                      
        if 'EOSLASER_CTL1_SL' not in existingFields:
            arcpy.AddField_management(feature_layer,
                                      'EOSLASER_CTL1_SL',
                                      field_type="FLOAT",
                                      field_alias='CP1 Laser Inclination',
                                      field_is_nullable="NULLABLE"
                                      )
                                      
        if 'EOSLASER_BS_LAT' not in existingFields:
            arcpy.AddField_management(feature_layer,
                                      'EOSLASER_BS_LAT',
                                      field_type="FLOAT",
                                      field_alias='Backsight Latitude',
                                      field_is_nullable="NULLABLE"
                                      )
                                      
        if 'EOSLASER_BS_LON' not in existingFields:
            arcpy.AddField_management(feature_layer,
                                      'EOSLASER_BS_LON',
                                      field_type="FLOAT",
                                      field_alias='Backsight Longitude',
                                      field_is_nullable="NULLABLE"
                                      )
                                      
        if 'EOSLASER_BS_ALT' not in existingFields:
            arcpy.AddField_management(feature_layer,
                                      'EOSLASER_BS_ALT',
                                      field_type="FLOAT",
                                      field_alias='Backsight Altitude',
                                      field_is_nullable="NULLABLE"
                                      )
                                      
        if 'EOSLASER_BS_HRMS' not in existingFields:
            arcpy.AddField_management(feature_layer,
                                      'EOSLASER_BS_HRMS',
                                      field_type="FLOAT",
                                      field_alias='Backsight HRMS',
                                      field_is_nullable="NULLABLE"
                                      )
                                      
        if 'EOSLASER_BS_SATS' not in existingFields:
            arcpy.AddField_management(feature_layer,
                                      'EOSLASER_BS_SATS',
                                      field_type="SHORT",
                                      field_alias='Backsight Number of Sats',
                                      field_is_nullable="NULLABLE"
                                      )
                                      
        if 'EOSLASER_BS_FIXTYPE' not in existingFields:
            arcpy.AddField_management(feature_layer,
                                      'EOSLASER_BS_FIXTYPE',
                                      field_type="SHORT",
                                      field_alias='Backsight Fix Type',
                                      field_is_nullable="NULLABLE",
                                      field_domain = "ESRI_FIX_TYPE_DOMAIN"                               
                                      )
                                      
        if 'EOSLASER_BS_AGE' not in existingFields:
            arcpy.AddField_management(feature_layer,
                                      'EOSLASER_BS_AGE',
                                      field_type="FLOAT",
                                      field_alias='Backsight Diff Age',
                                      field_is_nullable="NULLABLE"
                                      )
                                      
        if 'EOSLASER_BS_DIFFID' not in existingFields:
            arcpy.AddField_management(feature_layer,
                                      'EOSLASER_BS_DIFFID',
                                      field_type="STRING",
                                      field_length=4,
                                      field_alias='Backsight Station ID',
                                      field_is_nullable="NULLABLE"
                                      )
                                      
        if 'EOSLASER_BS_AVG' not in existingFields:
            arcpy.AddField_management(feature_layer,
                                      'EOSLASER_BS_AVG',
                                      field_type="SHORT",
                                      field_alias='Backsight Points Averaged',
                                      field_is_nullable="NULLABLE"
                                      )
                                      
        if 'EOSLASER_BS_SLDIST' not in existingFields:
            arcpy.AddField_management(feature_layer,
                                      'EOSLASER_BS_SLDIST',
                                      field_type="FLOAT",
                                      field_alias='Backsight Slope Distance',
                                      field_is_nullable="NULLABLE"
                                      )
                                      
        if 'EOSLASER_BS_AZI' not in existingFields:
            arcpy.AddField_management(feature_layer,
                                      'EOSLASER_BS_AZI',
                                      field_type="FLOAT",
                                      field_alias='Backsight Azimuth',
                                      field_is_nullable="NULLABLE"
                                      )
                                      
        if 'EOSLASER_BS_SL' not in existingFields:
            arcpy.AddField_management(feature_layer,
                                      'EOSLASER_BS_SL',
                                      field_type="FLOAT",
                                      field_alias='Backsight Inclination',
                                      field_is_nullable="NULLABLE"
                                      )
                                      
        if 'EOSLASER_BS_TRUEAZI' not in existingFields:
            arcpy.AddField_management(feature_layer,
                                      'EOSLASER_BS_TRUEAZI',
                                      field_type="FLOAT",
                                      field_alias='Computed True Azimuth',
                                      field_is_nullable="NULLABLE"
                                      )
                                      
        if 'EOSLASER_BS_AZICORR' not in existingFields:
            arcpy.AddField_management(feature_layer,
                                      'EOSLASER_BS_AZICORR',
                                      field_type="FLOAT",
                                      field_alias='Azimuth Correction',
                                      field_is_nullable="NULLABLE"
                                      )
                                      
        if 'EOSLASER_CTL2_LAT' not in existingFields:
            arcpy.AddField_management(feature_layer,
                                      'EOSLASER_CTL2_LAT',
                                      field_type="FLOAT",
                                      field_alias='CP2 Latitude',
                                      field_is_nullable="NULLABLE"
                                      )
                                      
        if 'EOSLASER_CTL2_LON' not in existingFields:
            arcpy.AddField_management(feature_layer,
                                      'EOSLASER_CTL2_LON',
                                      field_type="FLOAT",
                                      field_alias='CP2 Longitude',
                                      field_is_nullable="NULLABLE"
                                      )
                                      
        if 'EOSLASER_CTL2_ALT' not in existingFields:
            arcpy.AddField_management(feature_layer,
                                      'EOSLASER_CTL2_ALT',
                                      field_type="FLOAT",
                                      field_alias='CP2 Altitude',
                                      field_is_nullable="NULLABLE"
                                      )
                                      
        if 'EOSLASER_CTL2_HRMS' not in existingFields:
            arcpy.AddField_management(feature_layer,
                                      'EOSLASER_CTL2_HRMS',
                                      field_type="FLOAT",
                                      field_alias='CP2 HRMS',
                                      field_is_nullable="NULLABLE"
                                      )
                                      
        if 'EOSLASER_CTL2_SATS' not in existingFields:
            arcpy.AddField_management(feature_layer,
                                      'EOSLASER_CTL2_SATS',
                                      field_type="SHORT",
                                      field_alias='CP2 Number of Sats',
                                      field_is_nullable="NULLABLE"
                                      )
                                      
        if 'EOSLASER_CTL2_FIXTYPE' not in existingFields:
            arcpy.AddField_management(feature_layer,
                                      'EOSLASER_CTL2_FIXTYPE',
                                      field_type="SHORT",
                                      field_alias='CP2 Fix Type',
                                      field_is_nullable="NULLABLE",
                                      field_domain = "ESRI_FIX_TYPE_DOMAIN"
                                      )
                                      
        if 'EOSLASER_CTL2_AGE' not in existingFields:
            arcpy.AddField_management(feature_layer,
                                      'EOSLASER_CTL2_AGE',
                                      field_type="FLOAT",
                                      field_alias='CP2 Differential Age',
                                      field_is_nullable="NULLABLE"
                                      )
                                      
        if 'EOSLASER_CTL2_DIFFID' not in existingFields:
            arcpy.AddField_management(feature_layer,
                                      'EOSLASER_CTL2_DIFFID',
                                      field_type="STRING",
                                      field_length=4,
                                      field_alias='CP2 Station ID',
                                      field_is_nullable="NULLABLE"
                                      )
                                      
        if 'EOSLASER_CTL2_AVG' not in existingFields:
            arcpy.AddField_management(feature_layer,
                                      'EOSLASER_CTL2_AVG',
                                      field_type="SHORT",
                                      field_alias='CP2 Points Averaged',
                                      field_is_nullable="NULLABLE"
                                      )
                                      
        if 'EOSLASER_CTL2_SLDIST' not in existingFields:
            arcpy.AddField_management(feature_layer,
                                      'EOSLASER_CTL2_SLDIST',
                                      field_type="FLOAT",
                                      field_alias='CP2 Laser Slope Distance',
                                      field_is_nullable="NULLABLE"
                                      )
                                      
        if 'EOSLASER_CTL2_AZI' not in existingFields:
            arcpy.AddField_management(feature_layer,
                                      'EOSLASER_CTL2_AZI',
                                      field_type="FLOAT",
                                      field_alias='CP2 Laser Azimuth',
                                      field_is_nullable="NULLABLE"
                                      )
                                      
        if 'EOSLASER_CTL2_SL' not in existingFields:
            arcpy.AddField_management(feature_layer,
                                      'EOSLASER_CTL2_SL',
                                      field_type="FLOAT",
                                      field_alias='CP2 Laser Inclination',
                                      field_is_nullable="NULLABLE"
                                      )
        
        arcpy.AddMessage("Successfully added GPS Metadata fields.\n")

        # Update fields with Domains
        arcpy.AddMessage('Updating Fields with Domains...')
        
        # Update GNSS metadata fields with Domains
        domainFields = [field for field in arcpy.ListFields(feature_layer) if field.name == 'ESRIGNSS_FIXTYPE' or \
                          field.name == 'ESRIGNSS_STATIONID' or field.name == 'ESRIGNSS_NUMSATS']
        
        for field in domainFields:
            if field.name == 'ESRIGNSS_FIXTYPE' and not field.domain:
                arcpy.AssignDomainToField_management(feature_layer, field, 'ESRI_FIX_TYPE_DOMAIN')
                continue
            
            if field.name == 'ESRIGNSS_STATIONID' and not field.domain:
                arcpy.AssignDomainToField_management(feature_layer, field, 'ESRI_STATION_ID_DOMAIN')
                continue
            
            if field.name == 'ESRIGNSS_NUMSATS' and not field.domain:
                arcpy.AssignDomainToField_management(feature_layer, field, 'ESRI_NUM_SATS_DOMAIN')
                continue
            
        arcpy.AddMessage("Successfully updated GPS Metadata fields with domains.\n")
        
    except Exception as e:
        arcpy.AddError("{}\n".format(e))
        return

if __name__ == "__main__":
    """
        Commandline use to add fields to a layer

        Input: layer names (fully qualified paths)

        Example: python add_gps_fields "C:/temp/test.gdb/test" "C:/temp/test.gdb/test2"
    """
    parser = argparse.ArgumentParser("Add GPS Fields to Feature Layers")
    parser.add_argument("layers", nargs='+', help="The layers to add fields to")
    args = parser.parse_args()
    for layer in args.layers:
        add_gnss_fields(layer)
